import subprocess
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import altair as alt
import flatlatex
import numpy as np
import pandas as pd

Chart = alt.vegalite.v4.api.Chart


def display_img(img: str) -> None:
    """Display image inside notebook while preventing browser caching

    Args:
        img: Path to image
    """
    import IPython.display as IPd

    IPd.display(IPd.HTML('<img src="{}?now={}" / >'.format(img, time.time())))


def rgb2hex(r: int, g: int, b: int) -> str:
    """Convert RGB to HEX"""
    return "#{:02x}{:02x}{:02x}".format(r, g, b)


def hex2rgb(hex: str) -> Tuple[int]:
    """Convert HEX to RGB"""
    return tuple(map(ord, hex[1:].decode("hex")))


def convert_file(
    fn: str,
    to: str = "png",
    dpi: int = 300,
    background: Optional[str] = None,
    debug: bool = False,
) -> str:
    """Convert files with Inkscape

    Assumes that Inkscape can be called from shell with `inkscape`.

    Args:
        fn: Filename
        to: Target format
        dpi: Resolution
        background: Background color for `--export-background`
        debug: Debug mode

    Returns:
        New filename
    """
    fn = Path(fn).absolute()
    fn_new = fn.with_suffix(f".{to}")

    cmd = f"inkscape --export-filename={str(fn_new)} {str(fn)} --export-dpi={dpi}"

    if background is not None:
        cmd += " --export-background='{background}'"

    if debug:
        print(cmd)
    else:
        cmd += " &> /dev/null"

    subprocess.call(cmd, shell=True)

    return fn_new


def save(
    chart: Chart,
    path: str,
    extra_formats: Optional[Union[str, List[str]]] = None,
    debug: bool = False,
):
    """Save chart using `altair_saver`

    `altair_saver` seems to work best with `.svg` exports. When `extra_formats` is
    specified, the saved file (e.g., a `svg`) can be converted using the `convert_file`
    function which uses `inkscape` for file conversion (e.g., to save `png` or `pdf`
    versions of a chart).

    Args:
        chart: Chart to save
        path: Path to save chart to.
        extra_formats: Extra formats to save chart as, uses `convert_file` on saved file
    """
    try:
        import altair_saver
    except ImportError:
        print("altair_saver is required")
        print("http://github.com/altair-viz/altair_saver/")
        quit()

    path = str(path)
    chart.save(path)

    if extra_formats is not None:
        formats = (
            [extra_formats] if not isinstance(extra_formats, list) else extra_formats
        )
        for ff in formats:
            convert_file(path, to=ff, debug=debug)


def latex2unicode(latex: Union[str, List[str]]) -> Union[str, List[str]]:
    """Converts latex strings to unicode using `flatlatex`

    Args:
        latex: Latex strings (single string or list of strings)

    Returns:
        Unicode strings
    """
    c = flatlatex.converter()
    if type(latex) == str:
        return c.convert(latex)
    elif type(latex) == list:
        return [c.convert(entry) for entry in latex]
    else:
        raise NotImplementedError


def np2df(
    samples: Union[np.ndarray, List[np.ndarray]],
    field: str = "samples",
    labels_samples: Optional[Union[str, List[str]]] = None,
    labels_dim: List[str] = None,
    drop_dim: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Converts numpy arrays to pandas DataFrame used for plotting with `deneb.pairplot`

    Args:
        samples: Samples
        field: Field for sample identifier
        labels_samples: Labels for samples
        labels_dim: Labels for dimensions
        drop_dim: Dimensions to drop

    Returns:
        Formatted pandas DataFrame
    """
    if not isinstance(samples, list):
        samples = [samples]
    samples = [s for s in samples]
    dim_samples = samples[0].shape[1]
    for sample in samples:
        assert sample.shape[1] == dim_samples

    if labels_samples is None:
        labels_samples = [f"sample {i+1}" for i in range(len(samples))]
    if not isinstance(labels_samples, list):
        labels_samples = [labels_samples]
    assert len(labels_samples) == len(samples)

    if labels_dim is None:
        labels_dim = [f"dim {i+1}" for i in range(dim_samples)]
    assert len(labels_dim) == dim_samples

    dfs = []
    for i, sample in enumerate(samples):
        df = pd.DataFrame(sample, columns=labels_dim)
        df[field] = labels_samples[i]
        dfs.append(df)
    df = pd.concat(dfs, ignore_index=True)

    if drop_dim is not None:
        df = df.drop(columns=drop_dim)
        labels_dim = [ld for ld in labels_dim if ld not in drop_dim]

    return df


def colorscale(
    colors: Dict[str, str], shorthand: str = "sample:N", legend: bool = False,
) -> alt.Color:
    """Build colorscale

    Args:
        colors: Dictionary with domain as keys and range as values
        shorthand: Shorthand for color field
        legend: Whether to show a legend

    Returns:
        Colorscale
    """
    color_params = {
        "scale": alt.Scale(
            domain=[k for k in colors.keys()], range=[v for v in colors.values()]
        ),
    }

    if legend is False:
        color_params["legend"] = None
    else:
        color_params["legend"] = alt.Legend(title=None, orient="right", columns=1)

    colorscale = alt.Color(shorthand, **color_params)

    return colorscale



from typing import Any, Dict, List, Optional, Tuple, Union

import altair as alt
import pandas as pd

from .utils import Chart


def correlation_matrix(
    df: pd.DataFrame,
    metrics: List[str],
    limits: Optional[Union[Tuple[float], List[float]]] = None,
    font_size: int = 12,
    sparse: bool = True,
    white_font: float = 0.5,
    rotate_outwards: bool = False,
    legend: bool = False,
    height: Optional[int] = None,
    width: Optional[int] = None,
) -> Chart:
    """Correlation matrix

    Args:
        df: Dataframe
        metrics: Metrics to correlate (columns of dataframe)
        limits: Limits
        font_size: Font size for correlations
        sparse: Hides extra fields
        white_font: Threshold correlation for using white instead of black
        rotate_outwards: Rotates labels away from matrix
        legend: Whether or not to show legend
        height: Height of plot in facet
        width: Width of plot in facet

    Returns:
        Chart
    """
    valid_metrics = [m for m in metrics if m in df.columns]
    df_valid = df[valid_metrics]

    corrMatrix = df_valid.corr().reset_index().melt("index")
    corrMatrix.columns = ["var1", "var2", "correlation"]

    axis_params = {
        "domainWidth": 0,
        "tickWidth": 0,
    }

    if not rotate_outwards:
        xaxis = alt.Axis(labelAngle=0.0, **axis_params)
        yaxis = alt.Axis(labelAngle=270.0, **axis_params)
    else:
        xaxis = alt.Axis(labelAngle=270.0, **axis_params)
        yaxis = alt.Axis(labelAngle=0.0, **axis_params)

    color_params = {}
    if legend is False:
        color_params["legend"] = None
    else:
        color_params["legend"] = alt.Legend(title=None, orient="right", columns=1)
    if limits is not None:
        color_params["scale"] = alt.Scale(domain=limits)

    chart = (
        alt.Chart(corrMatrix)
        .mark_rect()
        .encode(
            x=alt.X("var1", title=None, axis=xaxis),
            y=alt.Y("var2", title=None, axis=yaxis),
            color=alt.Color("correlation", **color_params),
        )
    )

    chart += chart.mark_text(size=font_size).encode(
        text=alt.Text("correlation", format=".2f"),
        color=alt.condition(
            f"datum.correlation > {white_font}", alt.value("white"), alt.value("black"),
        ),
    )

    if sparse:
        chart = chart.transform_filter("datum.var1 < datum.var2")

    if height is not None:
        chart = chart.properties(height=height)

    if width is not None:
        chart = chart.properties(width=width)

    return chart



from typing import Any, Dict, List, Optional, Tuple, Union

import altair as alt
import pandas as pd

from .utils import Chart


def lineplot(
    df,
    x: str,
    y: str,
    label_orient: str = "top",
    title_orient: str = "bottom",
    column: Optional[str] = None,
    column_title: str = "",
    column_sort: Optional[List[str]] = None,
    column_labels: bool = True,
    row: Optional[str] = None,
    row_title: str = "",
    row_sort: Optional[List[str]] = None,
    row_labels: bool = True,
    aggregate: Optional[str] = "mean",
    points: bool = True,
    lines: bool = True,
    errorbars: bool = True,
    error_extent: str = "stdev",
    limits: Optional[List[float]] = None,
    independent_x: bool = True,
    independent_y: bool = False,
    spacing: int = 5,
    color: Optional[Union[str, alt.Color]] = None,
    shape: Optional[Union[str, alt.Shape]] = None,
    detail: Optional[Union[str, alt.Detail]] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
) -> Chart:
    """Produces lineplot with optional errorbars and facets

    Args:
        df: Dataframe
        x: Shorthand for x
        y: Shorthand for y
        label_orient: Orientation of column labels
        title_orient: Orientation of column title
        column: Shorthand for columns (optional) which are faceted
        column_title: Title text for columns displayed 
        column_sort: Optional sorting for columns
        column_labels: Whether header labels should be displayed
        row: Shorthand for rows (optional) which are faceted
        row_title: Title text for rows displayed 
        row_sort: Optional sorting for rows
        row_labels: Whether header labels should be displayed
        aggregate: Aggregation function for y values
        points: Whether to show points
        lines: Whether to show lines
        errorbars: Whether to show errorbars
        error_extent: Extent of errorbars, e.g., stdev or stderr
        limits: Limits for y-axis
        independent_x: Whether x-axes are independent
        independent_y: Whether y-axes are independent
        spacing: Spacing between facets
        color: Colorscale
        shape: Shape encoding
        detail: Detail encoding
        height: Height of plot in facet
        width: Width of plot in facet

    Returns:
        Chart
    """
    if color is None:
        color_kwarg = {}
    else:
        color_kwarg = {"color": color}

    if shape is None:
        shape_kwarg = {}
    else:
        shape_kwarg = {"shape": shape}

    if detail is None:
        detail_kwarg = {}
    else:
        detail_kwarg = {"detail": detail}

    if limits is not None:
        y_scale = alt.Scale(zero=False, domain=limits)
    else:
        y_scale = alt.Scale(zero=False)

    lines_layer = (
        alt.Chart()
        .mark_line()
        .encode(
            x=alt.X(x, title=""),
            y=alt.Y(y, scale=y_scale, aggregate=aggregate),
            **color_kwarg,
            **shape_kwarg,
            **detail_kwarg,
        )
    )

    points_layer = (
        alt.Chart()
        .mark_point(filled=True)
        .encode(
            x=alt.X(x, title=""),
            y=alt.Y(y, scale=y_scale, aggregate=aggregate),
            **color_kwarg,
            **shape_kwarg,
            **detail_kwarg,
        )
    )

    errorbars_layer = (
        alt.Chart()
        .mark_errorbar(extent=error_extent)
        .encode(
            x=alt.X(x, title=""),
            y=alt.Y(y, scale=y_scale),
            **color_kwarg,
            **shape_kwarg,
            **detail_kwarg,
        )
    )

    layers = []
    if lines:
        layers.append(lines_layer)
    if points:
        layers.append(points_layer)
    if errorbars:
        layers.append(errorbars_layer)

    chart = alt.layer(*layers, data=df)

    if height is not None:
        chart = chart.properties(height=height)

    if width is not None:
        chart = chart.properties(width=width)

    if column is not None or row is not None:
        facet_kwargs = {
            "spacing": spacing,
        }
        if column is not None:
            facet_kwargs["column"] = alt.Column(
                column,
                title=column_title,
                header=alt.Header(labels=column_labels),
                sort=column_sort,
            )
        if row is not None:
            facet_kwargs["row"] = alt.Row(
                row,
                title=row_title,
                header=alt.Header(labels=row_labels),
                sort=row_sort,
            )
        chart = chart.facet(**facet_kwargs)

    if independent_x:
        chart = chart.resolve_scale(x="independent")

    if independent_y:
        chart = chart.resolve_scale(y="independent")

    chart = chart.configure_header(titleOrient=title_orient, labelOrient=label_orient)

    return chart



from typing import Any, Dict, List, Optional, Tuple, Union

import altair as alt
import pandas as pd

from .utils import Chart


def pairplot(
    df: pd.DataFrame,
    field: str = "samples",
    limits: Optional[List[float]] = None,
    values: Optional[List[float]] = None,
    scatter_size: float = 1.5,
    legend: bool = False,
    format: Optional[str] = None,
    color: Optional[Union[str, alt.Color]] = None,
    num_bins: int = 20,
    bar_opacity: float = 0.8,
    clip: bool = True,
    height: Optional[int] = None,
    width: Optional[int] = None,
    debug: bool = False,
) -> Chart:
    """Pairplot

    Args:
        df: Samples as as pandas DataFrame. The dataframe needs one column
            labelled with the value of field (defaults to `sample`) which contains the
            sample identifier as a string. All other columns should contain float
            values which will be plotted. The utility function `deneb.np2df` may be
            used to turn a NumPy array into the correct format.
        limits: Limits for plotting (optional)
        values: Tickmark locations (optional)
        scatter_size: Size of scatter points
        legend: Whether or not to show legend
        format: Format string for axis ticks (optional)
        color: Colorscale, e.g. constructed with `deneb.colorscale`
        num_bins: Number of bins
        bar_opacity: Opacity of bars
        clip: Whether to clip
        height: Height of subplot
        width: Width of subplot
        debug: Debug mode

    Returns:
        Chart

    Note:
        Build on top of https://github.com/alt-viz/alt/issues/1926.
    """
    labels_dim = [c for c in list(df.columns) if c != field]
    assert len(list(df.columns)) == len(labels_dim) + 1

    if limits is not None:
        assert type(limits) == list
        if type(limits[0]) == float:
            assert len(limits) == 2
            limits_dim = {d: limits for d in labels_dim}
        else:
            assert type(limits[0]) == list
            assert len(limits) == dim_samples
            limits_dim = {d: limits[i] for i, d in enumerate(labels_dim)}

    axis_format = {"format": format} if format is not None else {}

    if color is None:
        color_kwarg = {}
    else:
        color_kwarg = {"color": color}

    def make_upper(r, w, no_x=True, no_y=True, **kwargs):
        enc_x_keywords = {"type": "quantitative"}
        enc_y_keywords = {"type": "quantitative"}

        if limits is not None:
            enc_x_keywords["scale"] = alt.Scale(domain=limits_dim[r], zero=False)
            enc_y_keywords["scale"] = alt.Scale(domain=limits_dim[w], zero=False)
            axis_limits = {"values": limits_dim[r]}
        else:
            axis_limits = {}

        if no_x:
            enc_x_keywords["axis"] = alt.Axis(
                title="", labels=False, domainWidth=0, tickWidth=0,
            )
        else:
            enc_x_keywords["axis"] = alt.Axis(**axis_limits, **axis_format,)
        if no_y:
            enc_y_keywords["axis"] = alt.Axis(
                title="", labels=False, domainWidth=0, tickWidth=0,
            )
        else:
            enc_y_keywords["axis"] = alt.Axis(
                title=w,
                labels=False,
                domainWidth=0,
                tickWidth=0,
                **axis_limits,
                **axis_format,
            )

        if debug:
            enc_x_keywords["axis"] = alt.Axis(title=r, **axis_limits, **axis_format,)
            enc_y_keywords["axis"] = alt.Axis(title=w, **axis_limits, **axis_format,)

        chart = (
            alt.Chart()
            .mark_circle(size=scatter_size, clip=clip)
            .encode(
                alt.X(r, **enc_x_keywords), alt.Y(w, **enc_y_keywords), **color_kwarg,
            )
        )

        if height is not None:
            chart = chart.properties(height=height)

        if width is not None:
            chart = chart.properties(width=width)

        return chart

    def make_lower(r, w, *args, **kwargs):
        return make_upper(r, w, *args, **kwargs)

    def make_diag(r, no_x=True, no_y=True, **kwargs):
        enc_x_keywords = {}
        enc_y_keywords = {}

        enc_x_keywords["type"] = "quantitative"

        if limits is not None:
            enc_x_keywords["scale"] = alt.Scale(domain=limits_dim[r], zero=False)
            axis_limits = {"values": limits_dim[r]}
        else:
            axis_limits = {}

        if no_x:
            enc_x_keywords["axis"] = alt.Axis(
                title="", labels=False, domainWidth=0, tickWidth=0,
            )
        else:
            enc_x_keywords["axis"] = alt.Axis(title=r, **axis_limits, **axis_format,)

        if no_y:
            enc_y_keywords["axis"] = alt.Axis(
                title="", labels=False, domainWidth=0, tickWidth=0,
            )
        else:
            enc_y_keywords["axis"] = alt.Axis(
                title=r, labels=False, domainWidth=0, tickWidth=0, **axis_format,
            )

        if debug:
            enc_x_keywords["axis"] = alt.Axis(title=r, **axis_limits, **axis_format)
            enc_y_keywords["axis"] = alt.Axis(title=r, labels=False, **axis_format)

        chart = alt.Chart()

        if limits is None:
            bin_keywords = {"maxbins": num_bins}
        else:
            bin_keywords = {
                "extent": limits_dim[r],
                "step": (limits_dim[r][1] - limits_dim[r][0]) / num_bins,
            }

        chart = chart.mark_bar(opacity=bar_opacity, clip=clip)
        chart = chart.encode(
            x=alt.X(r, bin=alt.Bin(**bin_keywords), **enc_x_keywords,),
            y=alt.Y("count()", stack=None, **enc_y_keywords),
            **color_kwarg,
        )

        if height is not None:
            chart = chart.properties(height=height)

        if width is not None:
            chart = chart.properties(width=width)

        return chart

    instance = (
        Matrix(df)
        .fields(*labels_dim)
        .upper(make_upper)
        .diag(make_diag)
        .lower(make_lower)
    )

    chart = instance.render()

    return chart


class Matrix:
    def __init__(self, data):
        self.data = data
        self.cols = []
        self.rows = []
        self.grid = []

    def fields(self, *args):
        self.cols = list(args)
        self.rows = list(args)
        self.grid = [[None for _ in self.rows] for _ in self.cols]
        return self

    def diag(self, make_chart):
        for i, r in enumerate(self.rows):
            no_x = False if i == (len(self.rows) - 1) else True
            no_y = False if i == 0 else True
            self.grid[i][i] = make_chart(r, no_x=no_x, no_y=no_y)

        return self

    def lower(self, make_chart):
        for i, r in enumerate(self.rows):
            for j, c in enumerate(self.rows):
                if i > j:
                    no_x = False if i == (len(self.rows) - 1) else True
                    no_y = False if j == 0 else True
                    self.grid[i][j] = make_chart(c, r, no_x=no_x, no_y=no_y)

        return self

    def upper(self, make_chart):
        for i, r in enumerate(self.rows):
            for j, c in enumerate(self.cols):
                if j > i:
                    no_x = no_y = True
                    self.grid[i][j] = make_chart(c, r, no_x=no_x, no_y=no_y)
        return self

    def render(self):
        chart = alt.vconcat(data=self.data)
        for i, _ in enumerate(self.rows):
            row = alt.hconcat()
            for j, _ in enumerate(self.cols):
                if self.grid[i][j] is not None:
                    row |= self.grid[i][j]
            if len(row.hconcat) > 0:
                chart &= row

        return chart
